;; cl-chmlib - Bindings
;; Copyright (C) 2006  Frederic Jolliton <frederic@jolliton.com>
;; 
;; This library is free software; you can redistribute it and/or
;; modify it under the terms of the GNU Lesser General Public
;; License as published by the Free Software Foundation; either
;; version 2.1 of the License, or (at your option) any later version.
;; 
;; This library is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;; Lesser General Public License for more details.
;; 
;; You should have received a copy of the GNU Lesser General Public
;; License along with this library; if not, write to the Free Software
;; Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

(in-package #:net.tuxee.chmlib)

(defun foreign-u8-array-to-lisp (ptr count)
  (let ((result (make-array count :element-type '(unsigned-byte 8))))
    (loop :for i :below count
          :do (setf (aref result i) (mem-ref ptr :unsigned-char i)))
    result))

(defvar *chm-enumerator-callback* nil)

(defconstant +chm-flag-normal+ 1)
(defconstant +chm-flag-meta+ 2)
(defconstant +chm-flag-special+ 4)
(defconstant +chm-flag-files+ 8)
(defconstant +chm-flag-dirs+ 16)
(defconstant +chm-flag-all+ 31)

(defconstant +chm-enumerator-failure+ 0)
(defconstant +chm-enumerator-continue+ 1)
(defconstant +chm-enumerator-success+ 2)

(define-foreign-library libchm
  (t (:default "libchm")))

(use-foreign-library libchm)

(defcfun ("chm_open" c-chm-open) :pointer (filename :string))

(defcfun ("chm_close" c-chm-close) :void (handle :pointer))

(defcstruct s-chm-unit-info
  (start :uint64)
  (length :uint64)
  (space :int)
  (flags :int)
  (path :char :count 513))

(defcfun ("chm_resolve_object" c-chm-resolve-object)
    :int (chm :pointer) (path :string) (unit-info :pointer))

(defcfun ("chm_retrieve_object" c-chm-retrieve-object)
    :uint64
    (chm :pointer) (unit-info :pointer) (buffer :pointer)
    (offset :uint64) (length :int64))

(defun chm-extract-unit-info (unit-info)
  (with-foreign-slots ((start length space flags path)
                       unit-info s-chm-unit-info)
    `((:path . ,(foreign-string-to-lisp path))
      (:start . ,start)
      (:length . ,length)
      (:space . ,space)
      (:flags . ,flags))))

;; Warning: Unspecified behavior !
;;
;; However, it *seems* to work on Linux 2.6 (x86-32), with ACL 7.0,
;; ACL 8.0, LW 4.4, SBCL 0.9.13, CMUCL 19c, CLISP 2.38, GCL 2.6.7,
;; ECL 0.9f and ABCL 0.0.9.
(defmacro force-local-exit (value &body body)
  `(block normal-exit
     (let (result result-p)
       (unwind-protect (setf result (progn ,@body)
                             result-p t)
         (return-from normal-exit (if result-p result ,value))))))

(defcallback chm-enumerator
    :int ((chm :pointer) (unit-info :pointer) (context :pointer))
  (declare (ignore chm context))
  (force-local-exit +chm-enumerator-failure+
    (if (funcall *chm-enumerator-callback* (chm-extract-unit-info unit-info))
        +chm-enumerator-continue+
        +chm-enumerator-failure+)))

(defcfun ("chm_enumerate" c-chm-enumerate)
    :int (chm :pointer) (what :int) (callback :pointer) (context :pointer))

(defun chm-enumerate (chm what callback)
  "Call CALLBACK for each entry in CHM matching WHAT flags."
  (let ((*chm-enumerator-callback* callback))
    (c-chm-enumerate chm what (callback chm-enumerator) (null-pointer)))
  (values))

(defcfun ("chm_enumerate_dir" c-chm-enumerate-dir)
    :int (chm :pointer) (prefix :string) (what :int) (callback :pointer) (context :pointer))

(defun chm-enumerate-directory (chm prefix what callback)
  "Call CALLBACK for each entry in CHM with prefix PREFIX
  matching WHAT flags."
  (let ((*chm-enumerator-callback* callback))
    (c-chm-enumerate-dir chm prefix what (callback chm-enumerator) (null-pointer)))
  (values))

(defmacro with-chm-unit-info ((unit-info chm path) &body body)
  (let ((%chm (gensym))
        (%path (gensym))
        (%unit-info (gensym)))
    `(let ((,%chm ,chm)
           (,%path ,path))
       (with-foreign-object (,%unit-info 's-chm-unit-info)
         (setf (foreign-slot-value ,%unit-info 's-chm-unit-info 'flags) 0)
         (unless (zerop (c-chm-resolve-object ,%chm ,%path ,%unit-info))
           (error "Object ~S not found" ,%path))
         (let ((,unit-info ,%unit-info))
           ,@body)))))

(defun chm-object-info (chm path)
  "Retrieve object information"
  (with-chm-unit-info (unit-info chm path)
    (chm-extract-unit-info unit-info)))

(defun chm-retrieve-object (chm path &key (offset 0) (length nil))
  "Retrieve object contents"
  (with-chm-unit-info (unit-info chm path)
    (let ((object-length (foreign-slot-value unit-info 's-chm-unit-info 'length)))
      (with-foreign-object (buffer :char object-length)
        (let ((effective-length
               (c-chm-retrieve-object chm unit-info buffer
                                      offset (or length object-length))))
          (foreign-u8-array-to-lisp buffer effective-length))))))

(defmacro with-chm ((var filename) &body body)
  "Open a chm from FILENAME, then execute body and take care to
  close the chm."
  (let ((%chm (gensym))
        (%filename (gensym)))
    `(progn
       (let* ((,%filename (namestring ,filename))
              (,%chm (c-chm-open ,%filename)))
         (when (null-pointer-p ,%chm)
           (error "Error while opening ~S" ,%filename))
         (unwind-protect
              (let ((,var ,%chm))
                ,@body)
           (c-chm-close ,%chm))))))
